﻿using System;
using FCSColl = FCSlib.Data.Collections;

namespace ThirdLeftChild {
  class MainClass {
    static void Main(string[] args) {

      AccessTreeStep1( );
//      AccessTreeStep2( );
//      AccessTreeStep3( );

//      SelectManyBinding();
    }

    #region Step 1
    private static void AccessTreeStep1( ) {
      var tree = new FCSColl::UnbalancedBinaryTree<string>( );
      tree = tree.Insert("Paul");
      tree = tree.Insert("Adam");
      tree = tree.Insert("Bernie");
      tree = tree.Insert("Willy");
      tree = tree.Insert("Suzie");
      Console.WriteLine(tree);

      // Try to get hold of the third "left" child from the top node.
      // This throws an exception, because tree.Left.Left.Left is null
      //Console.WriteLine(tree.Left.Left.Left.Value);

      // Problem: when accessing elements in a chain like this, any one of them
      // could be null. So, to be safe:
      Console.WriteLine(GetThirdLeftChild(tree));

      // Alternatively, in C# 6.0, you could use these variations with 
      // the null-conditional operator or even combining the 
      // null-conditional and the null-coalescing operators
      Console.WriteLine(GetThirdLeftChildNullConditional(tree));
      Console.WriteLine(GetThirdLeftChildNullConditional_(tree));
    }


    static string GetThirdLeftChild(FCSColl::UnbalancedBinaryTree<string> tree) {
      if (tree != null) {
        if (tree.Left != null) {
          if (tree.Left.Left != null) {
            if (tree.Left.Left.Left != null) {
              return tree.Left.Left.Left.Value;
            }
            else {
              return "No such child";
            }
          }
          else {
            return "No such child";
          }
        }
        else {
          return "No such child";
        }
      }
      else {
        return "No such child";
      }
    }

    static string GetThirdLeftChildNullConditional(
      FCSColl::UnbalancedBinaryTree<string> tree) =>
      tree?.Left?.Left?.Left?.Value;

    static string GetThirdLeftChildNullConditional_(
      FCSColl::UnbalancedBinaryTree<string> tree) =>
      tree?.Left?.Left?.Left?.Value  ?? "No such child";

    #endregion

    #region Step 2
    private static void AccessTreeStep2( ) {
      var tree = new FCSColl::UnbalancedBinaryTree<string>( );
      tree = tree.Insert("Paul");
      tree = tree.Insert("Adam");
      tree = tree.Insert("Bernie");
      tree = tree.Insert("Willy");
      tree = tree.Insert("Suzie");
      Console.WriteLine(tree);

      var thirdLeftChild = GetMaybeThirdLeftChild(tree);
      if (thirdLeftChild.HasValue)
        Console.WriteLine(thirdLeftChild.Value);
      else
        Console.WriteLine("Maybe is empty");
    }

    static Maybe<string> GetMaybeThirdLeftChild(FCSColl::UnbalancedBinaryTree<string> tree) {
      if (tree != null) {
        if (tree.Left != null) {
          if (tree.Left.Left != null) {
            if (tree.Left.Left.Left != null) {
              return tree.Left.Left.Left.Value.ToMaybe( );
            }
            else {
              return Maybe<string>.Empty;
            }
          }
          else {
            return Maybe<string>.Empty;
          }
        }
        else {
          return Maybe<string>.Empty;
        }
      }
      else {
        return Maybe<string>.Empty;
      }
    }

    #endregion

    #region Step 3
    private static void AccessTreeStep3( ) {
      var tree = new FCSColl::UnbalancedBinaryTree<string>( );
      tree = tree.Insert("Paul");
      tree = tree.Insert("Adam");
      tree = tree.Insert("Bernie");
      tree = tree.Insert("Willy");
      tree = tree.Insert("Suzie");
      Console.WriteLine(tree);

      var thirdLeftChild = GetMonadicThirdLeftChild(tree);
      if (thirdLeftChild.HasValue)
        Console.WriteLine(thirdLeftChild.Value.Value);
      else
        Console.WriteLine("Maybe is empty again");

      // Now we've seen all the failures, here's the same for the child one left one right from the top,
      // which should be Bernie:
      var bernie =
        tree.ToNotNullMaybe( ).
        Bind(t => t.Left.ToNotNullMaybe( )).
        Bind(t => t.Right.ToNotNullMaybe( ));
      Console.WriteLine(bernie.HasValue ? bernie.Value.Value : "No Bernie found");
    }

    static Maybe<FCSColl::UnbalancedBinaryTree<string>> 
    GetMonadicThirdLeftChild(FCSColl::UnbalancedBinaryTree<string> tree) {
      return tree.ToNotNullMaybe( ).
        Bind(t => t.Left.ToNotNullMaybe( )).
        Bind(t => t.Left.ToNotNullMaybe( )).
        Bind(t => t.Left.ToNotNullMaybe( ));
    }

    // Alternative syntax using fancy operator overloading
    static Maybe<FCSColl::UnbalancedBinaryTree<string>> GetMonadicThirdLeftChild_(FCSColl::UnbalancedBinaryTree<string> tree) {
      return tree.ToNotNullMaybe( ) &
        (t => t.Left.ToNotNullMaybe( )) &
        (t => t.Left.ToNotNullMaybe( )) &
        (t => t.Left.ToNotNullMaybe( ));
    }   

    // This is for the idea (see Maybe.cs) of shifting the 
    // ToNotNullMaybe call into the monad.

    //    static Maybe<FCSColl::UnbalancedBinaryTree<string>> GetMonadicThirdLeftChild__(FCSColl::UnbalancedBinaryTree<string> tree) {
    //      return tree.ToNotNullMaybe( ) &
    //        (t => t.Left) &
    //        (t => t.Left) &
    //        (t => t.Left);
    //    }
    #endregion

    #region SelectManyBinding
    static void SelectManyBinding() {
      // Another simple calculation using Bind
      var result1 = 5.ToMaybe().
        Bind(v => Maybe<int>.Empty.Bind(v2 => (v + v2).ToMaybe()));
      Console.WriteLine("Result 1: " + (result1.HasValue ? result1.Value.ToString() : "None"));

      // The same thing can be done using SelectMany
      // implemented for the Maybe monad with the standard
      // signature
      var result2 = 5.ToMaybe().SelectMany(
                      v => Maybe<int>.Empty.SelectMany(
                        v2 => (v + v2).ToMaybe()));
      Console.WriteLine("Result 2: " + (result2.HasValue ? result2.Value.ToString() : "None"));

      // Using the second SelectMany overload, which is used 
      // by standard LINQ de-sugaring, it is now possible
      // to take advantage of the Monad functionality by
      // performing calculations in LINQ expressions
      var result3 = 
        from x in 5.ToMaybe()
            from y in Maybe<int>.Empty
            select (x + y);
      Console.WriteLine("Result 3: " + (result3.HasValue ? result3.Value.ToString() : "None"));

      var result4 =
        from x in Maybe<int>.Empty
            from y in 5.ToMaybe()
            select (x + y);
      Console.WriteLine("Result 4: " + (result4.HasValue ? result4.Value.ToString() : "None"));

      var result5 =
        from x in 5.ToMaybe()
            from y in 7.ToMaybe()
            select (x + y);
      Console.WriteLine("Result 5: " + (result5.HasValue ? result5.Value.ToString() : "None"));
    }
    #endregion
  }
}
